#Exercise 3 template for CHE210D

#import python modules
import numpy as np
import random
import matplotlib.pyplot as plt
import time
import pickle

#import compiled fortran library
import finallib

import sys
import os

#whether or not to use the 3d visualization
UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
np.random.seed(4323450)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finallib.calcenergy(Pos + NormDir * x, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finallib.calcenergy(Pos + NormDir * dx, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finallib.calcenergy(Pos + NormDir * 2. * dx, L, Cut)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finallib.calcenergy(Pos + NormDir * Dists[-1], L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finallib.calcenergy(Pos + NormDir * d3, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finallib.calcenergy(Pos + NormDir * d3, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin,L)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin,L)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global L, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finallib.calcenergyforces(Pos, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = finallib.calcenergyforces(Pos, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.)+1)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T): #is this right?
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel) #kcal/mol
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    #Vel = np.random.rand(N,3)
    Vel = np.random.normal(0,T,size=N*3).reshape(N,3)    
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = finallib.calcenergyforces(Pos, L, Cut, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

#equilibrate with velocity rescaling
def Equilibrate(T, NstepsEquil):

    global L, N, Cut
    InitPos = InitPositions(N,L)
    PE, Pos = ConjugateGradient(InitPos,0.001,1.e-8,1.e-10)
    Vel = InitVelocities(N,T)
    Accel = InitAccel(Pos)

    t = 0
    RescaleFreq = 10
    dt = .001
    Nsteps = 0

    PE_VV = []
    tlist = []
#equilibrate with velocity rescaling
    results = file('PEs_equil1.txt',"w")
    results.write('t PE \n')

    while Nsteps <= NstepsEquil:
        Pos, Vel, Accel, KE, PE = finallib.vvintegrate(Pos, Vel, Accel, L, Cut, dt)
        tlist.extend([t])
        PE_VV.extend([PE])
        if t/dt % RescaleFreq == 0:
            Vel = RescaleVelocities(Vel,T)
    
        s = str(t)+' '+str(PE)+' '+'\n'
        results.write(s)
        t = t + dt
        Nsteps = Nsteps+1
        
      #  if UseVisual:
       #     if atomvis.Initialized:
        #       atomvis.Update(Pos,L)
	#else:
            #   atomvis.Init(Pos,L)
               
    results.close()

#second equilibration with velocity rescaling
    Nsteps = 0
    E = []
    U = []

    results = file('PEs_equil2.txt',"w")
    results.write('t PE E \n')

    while Nsteps <= NstepsEquil:
        Pos, Vel, Accel, KE, PE = finallib.vvintegrate(Pos, Vel, Accel, L, Cut, dt)
        tlist.extend([t])
        U.extend([PE])
        E.extend([PE+KE])
        if t/dt % RescaleFreq == 0:
            Vel = RescaleVelocities(Vel,T)
    
        s = str(t)+' '+str(PE)+' '+str(PE+KE)+'\n'
        results.write(s)
        t = t + dt
        Nsteps = Nsteps+1
        if UseVisual:
            if atomvis.Initialized:
                atomvis.Update(Pos,L)
            else:
                atomvis.Init(Pos,L)
        
    Eavg = np.mean(np.array([E]))
    newKE = Eavg - PE
    Vel = RescaleVelocities(Vel,newKE/(1.5*len(Vel)))
    results.close()
    
    return Pos, Vel, Accel

#production run
N = 216
rho = 0.48
L = (N/rho)**0.3333333333  #calculated from dimless rho = 0.48
Cut =2.5 #calculated from 2.5*sigma
T = 0.12 #dimensionless T
NStepsEquil = 10000


Pos, Vel, Accel = Equilibrate(T,NStepsEquil)
Pos0 = np.array(Pos.copy())

Nsteps = 0.0
NstepsProd = 100000.0
RecordFreq = 100.0

tlist = []
MSD = []
MSD_2D = []


results = file('production.txt',"w")
results.write('t MSD MSD_2D\n')

dt = .001

numbins = 500
delta = L/numbins
csum = np.zeros(numbins)
csum_wall = np.zeros(numbins)


while Nsteps <= NstepsProd:
    Pos, Vel, Accel, KE, PE = finallib.vvintegrate(Pos, Vel, Accel, L, Cut, dt)
    print  Nsteps/NstepsProd
    if UseVisual:
        if atomvis.Initialized:
            atomvis.Update(Pos,L)
        else:
            atomvis.Init(Pos,L)
    
   
    if Nsteps % RecordFreq == 0.0:
        tlist.extend([Nsteps*dt])
        results.write(str(Nsteps*dt)+' ')
        dis = Pos-Pos0
        MSD.extend([np.sum(dis*dis)])
        results.write(str(np.sum(dis*dis))+' ')

        #calculate MSD in x and z directions
        dis[:,1] = 0 #erases the y values 
        MSD_2D.extend([np.sum(dis*dis)])
        results.write(str(np.sum(dis*dis))+'\n')

        #count for g(r)        
        n = 0
        dlist = []
        while n < len(Pos)-1:
            r = Pos[n] - Pos[n + 1]
            r2 = np.sum(r*r)
            dlist.extend([np.sqrt(r2)])
            n = n + 1
            
        clist =[]
        i = 0
        lowbound = 0
        while i < numbins:
            hibound = (i + 1)*delta
            c = 0
            for element in dlist:
                if element >= lowbound and element < hibound:
                    c = c + 1
            clist.extend([c])
            lowbound = hibound
            i = i + 1
        csum = csum + np.array(clist)
        
        #count for distance from wall


        dis_wall = -L/2 - Pos
        dlist_wall = abs(dis_wall[:,1])
        
        cwall_list =[]
        i = 0
        lowbound = 0
        while i < numbins:
            hibound = (i + 1)*delta
            c = 0
            for element in dlist_wall:
                if element >= lowbound and element < hibound:
                    c = c + 1
            cwall_list.extend([c])
            lowbound = hibound
            i = i + 1
        csum_wall = csum_wall + np.array(cwall_list)
    
    Nsteps = Nsteps+1        

    
results.close()

    
# find g(r)
results = file('gcalc.txt',"w")
results.write('r g g_wall \n')   
rd = []
g = np.zeros(numbins)
g_wall = np.zeros(numbins)
dens = np.zeros(numbins)

i = 0
nobs = NstepsProd/RecordFreq

while i < numbins:
    rd.extend([i*delta])
    g[i] = csum[i]*(L/2)**3/(nobs*N*(N-1)/2*delta**3*((i + 1)**3-i**3))*1000
    g_wall[i] = csum_wall[i]*(L/2)**3/(nobs*N*delta**3*((i + 1)**3-i**3))
    results.write(str(i*delta)+' '+ str(g[i])+' ' +str(g_wall[i])+' ' +str(csum[i]/(nobs*N*(N-1)/2))+'\n')
    i = i + 1
    
results.close()


	