#Exercise 3 template for CHE210D

#import python modules
import numpy as np
import random
import matplotlib.pyplot as plt
import time
import pickle
import atomwrite

#import compiled fortran library
import finallib_wall

import sys
import os
import matplotlib.pyplot as plt

#whether or not to use the 3d visualization
UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions

#set the random number seed; useful for debugging
np.random.seed(4232234)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global L, Y
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finallib_wall.calcenergy(Pos + NormDir * x, L, Y) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finallib_wall.calcenergy(Pos + NormDir * dx, L, Y)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finallib_wall.calcenergy(Pos + NormDir * 2. * dx, L, Y)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finallib_wall.calcenergy(Pos + NormDir * Dists[-1], L, Y)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finallib_wall.calcenergy(Pos + NormDir * d3, L, Y)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finallib_wall.calcenergy(Pos + NormDir * d3, L, Y)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin,L)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin,L)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global L, Y
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finallib_wall.calcenergyforces(Pos, L, Y, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = finallib_wall.calcenergyforces(Pos, L, Y, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.) #get rid of the '1' ?
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
                    
    return Pos
    
def InitWall(N, L, Y):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.))
    #make the position array
    Pos = np.zeros((NLat*NLat,3), float)  
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) -0.5*L  
    bound = Y/2
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
                Pos[i] = np.array([x,bound,y], float)
                i += 1
                #if done placing atoms, return
                if i >= NLat*NLat:
                    a = -Pos
                    a[:,1] = Pos[:,1]
                    morePos = Pos.tolist()
                    morePos.extend(a.tolist())
                    a[:,0] = Pos[:,0]
                    morePos.extend(a.tolist())
                    a = -a
                    a[:,1] = Pos[:,1]
                    morePos.extend(a.tolist())
                    Pos = np.array(morePos)
    return Pos 


def RescaleVelocities(Vel, T): #is this right?
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel) #kcal/mol
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    #Vel = np.random.rand(N,3)
    Vel = np.random.normal(0,T,size=N*3).reshape(N,3)    
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global L
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = finallib_wall.calcenergyforces(Pos, L, Y, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

#equilibrate with velocity rescaling
def Equilibrate(T, NstepsEquil):

    global L, Y, N
    InitPos = InitPositions(N,L*0.6)
    #InitWall(N,L-2)
    PE, Pos = ConjugateGradient(InitPos,0.001,1.e-8,1.e-10)
    Vel = InitVelocities(N,T)
    Accel = InitAccel(Pos)

    t = 0
    RescaleFreq = 10
    dt = .001
    Nsteps = 0

    PE_VV = []
    tlist = []
#equilibrate with velocity rescaling
    results = file('PEs_equil1_wall.txt',"w")
    results.write('t PE \n')

    while Nsteps <= NstepsEquil:
        Pos, Vel, Accel, KE, PE = finallib_wall.vvintegrate(Pos, Vel, Accel, L, Y, dt)
        tlist.extend([t])
        PE_VV.extend([PE])
        if t/dt % RescaleFreq == 0:
            Vel = RescaleVelocities(Vel,T)
    
        s = str(t)+' '+str(PE)+' '+'\n'
        results.write(s)
        t = t + dt
        Nsteps = Nsteps+1
        if UseVisual:
            if atomvis.Initialized:
               atomvis.Update(Pos,L)
            else:
               atomvis.Init(Pos,L)
    results.close()

#second equilibration with velocity rescaling
    Nsteps = 0
    E = []
    U = []

    results = file('PEs_equil2_wall.txt',"w")
    results.write('t PE E \n')

    while Nsteps <= NstepsEquil:
        Pos, Vel, Accel, KE, PE = finallib_wall.vvintegrate(Pos, Vel, Accel, L, Y, dt)
        tlist.extend([t])
        U.extend([PE])
        E.extend([PE+KE])
        if t/dt % RescaleFreq == 0:
            Vel = RescaleVelocities(Vel,T)
    
        s = str(t)+' '+str(PE)+' '+str(PE+KE)+'\n'
        results.write(s)
        t = t + dt
        Nsteps = Nsteps+1
        if UseVisual:
            if atomvis.Initialized:
                atomvis.Update(Pos,L)
            else:
                atomvis.Init(Pos,L)
        
    Eavg = np.mean(np.array([E]))
    newKE = Eavg - PE
    Vel = RescaleVelocities(Vel,newKE/(1.5*len(Vel)))
    results.close()
    
    return Pos, Vel, Accel
#production run
N = 216
rho = 0.6
L = (N/rho)**0.33333/0.6 #multiplied by some factor such that initially they are at rho
Y =  0.8*L #this is the space between the hydrophobic plates

#(N/rho)**0.3333333333  #calculated from dimless rho = 0.48

T = 0.12 #dimensionless T
NStepsEquil = 10000


Pos, Vel, Accel = Equilibrate(T,NStepsEquil)
Pos0 = np.array(Pos.copy())

dis_bwall_Pos0 = -Y/2-Pos
dis_twall_Pos0 = Y/2-Pos
dlist_bwall_Pos0 = abs(dis_bwall_Pos0[:,1])
dlist_twall_Pos0 = abs(dis_twall_Pos0[:,1])

firstlayer = []
secondlayer = []
thirdlayer = []
bulk = []


n = 0
while n < len(Pos0)-1:
    if dlist_bwall_Pos0[n] <= 2 or dlist_twall_Pos0[n] <= 2:
        firstlayer.extend([n])
    else:
        if dlist_bwall_Pos0[n] <= 3 or dlist_twall_Pos0[n] <= 3:
            secondlayer.extend([n])
        else:
            if dlist_bwall_Pos0[n] <= 4 or dlist_twall_Pos0[n] <= 4:
                thirdlayer.extend([n])
            else:
                bulk.extend([n])
    n = n + 1
    
firstlayer = np.array(firstlayer)
secondlayer = np.array(secondlayer)
thirdlayer = np.array(thirdlayer)
bulk = np.array(bulk)


Nsteps = 0.0
NstepsProd = 100000.0
tlist = []
MSD = []

results = file('production_wall.txt',"w")
results.write('t MSD \n')

t = 0
dt = .001

numbins = 800 #i added more bins because the box is so big now
delta = L/numbins
deltay = Y/numbins
RecordFreq = 100
csum = np.zeros(numbins)
csum_wall = np.zeros(numbins)
sum_y = np.zeros(numbins)
sum_negy = np.zeros(numbins)



PosW1 = InitWall(N,L,Y)
PosW2 = InitWall(N,L,-Y)

PosWall = PosW1.tolist() + PosW2.tolist()

#Names = []
#n=0

#while n < len(PosW1)+len(PosW2)-1:
 #   Names.extend(['w'])
  #  n = n + 1

#=0
#while n < N:
 #   Names.extend(['H2O'])
  #  n = n+1
    

Pdb = atomwrite.pdbfile("water_anim.pdb",L)

Pos_write = PosWall + Pos.tolist()

Pdb.write(np.array(Pos_write))


#this is where MD simulation starts!!!

while Nsteps <= NstepsProd:
    Pos, Vel, Accel, KE, PE = finallib_wall.vvintegrate(Pos, Vel, Accel, L, Y, dt)
    print Nsteps/NstepsProd
    
    if UseVisual:
        if atomvis.Initialized:
            atomvis.Update(Pos,L)
        else:
            atomvis.Init(Pos,L)
    
   
    if Nsteps % RecordFreq == 0:

        #Pdb.write(PosW1)
        #Pdb.write(PosW2)
        Pos_write = PosWall + Pos.tolist()
        Pdb.write(np.array(Pos_write))
        
        n = 0
        dlist = []
        while n < len(Pos)-1:
            r = Pos[n] - Pos[n + 1]
            r2 = np.sum(r*r)
            dlist.extend([np.sqrt(r2)])
            n = n + 1
            
        clist =[]
        i = 0
        lowbound = 0
      
        while i < numbins:
            hibound = (i + 1)*delta
            c = 0
            for element in dlist:
                if element >= lowbound and element < hibound:
                    c = c + 1
            clist.extend([c])
            lowbound = hibound
            i = i + 1
            
        csum = csum + np.array(clist)
        
#find the distances of all the particles from the bottom wall
        #wall_d = Pos.copy()
        #wall_d[:,1] = -Y/2
        #dis_wall = Pos - wall_d
        #diswall2 = dis_wall*dis_wall
        #dlist_wall = np.sum(diswall2,axis=1)**0.5
            
        dis_wall = -L/2-Pos
        dlist_wall = abs(dis_wall[:,1])
        
        cwall_list =[]

        i = 0
        lowbound = 0
        deltalist=[]
        
        while i < numbins:
            deltalist.extend([lowbound])
            hibound = (i + 1)*delta
            c = 0
            for element in dlist_wall:
                if element >= lowbound and element < hibound:
                    c = c + 1
            cwall_list.extend([c])
            lowbound = hibound
            i = i + 1
        
           
        csum_wall = csum_wall + np.array(cwall_list)
        

        
 #calculate MSD in x and z directions
        tlist.extend([Nsteps*dt])
        results.write(str(Nsteps*dt)+' ')
        dis = Pos-Pos0 #i changed Pos from an array
        dis[:,1] = 0 #erases the y values 
        MSD.extend([np.sum(dis*dis)])
        results.write(str(np.sum(dis*dis))+' ')
        
        dlayer1 = []       
        for element in firstlayer:
            dlayer1.append((Pos[element] - Pos0[element]).tolist())
        dlayer1= np.array(dlayer1)
        dlayer1[:,1] = 0        
        results.write(str(np.sum(dlayer1*dlayer1))+ ' ')
        
        dlayer2 = []       
        for element in secondlayer:
            dlayer2.append((Pos[element] - Pos0[element]).tolist())
        dlayer2= np.array(dlayer2)
        dlayer2[:,1] = 0       
        results.write(str(np.sum(dlayer2*dlayer2))+ ' ')
        
        dlayer3 = []       
        for element in thirdlayer:
            dlayer3.append((Pos[element] - Pos0[element]).tolist())
        dlayer3= np.array(dlayer3)
        dlayer3[:,1] = 0       
        results.write(str(np.sum(dlayer3*dlayer3))+ ' ')
        
        if len(bulk) > 0:
            results.write(str(np.sum(dlayer3*dlayer3))+ ' ')
            dbulk = []       
            for element in bulk:
                dbulk.append((Pos[element] - Pos0[element]).tolist())
            dbulk= np.array(dbulk)
            dbulk[:,1] = 0    
            results.write(str(np.sum(dbulk*dbulk))+ '\n')
        else:
            results.write(str(np.sum(dlayer3*dlayer3))+ '\n')
             
 
    Nsteps = Nsteps+1
    
results.close()


Pdb.close()

plt.plot(np.array(deltalist),csum_wall)
    
    # find g(r)
results = file('gcalc_wall.txt',"w")
results.write('r g g_wall\n')   
rd = []
g = np.zeros(numbins)
g_wall = np.zeros(numbins)
dens = np.zeros(numbins)
i = 0
nobs = NstepsProd/RecordFreq

while i < numbins:
    rd.extend([i*delta])
    g[i] = 2*csum[i]/(nobs*N*(N-1))*L**3/(4*np.pi*delta**3/3*((i + 1)**3-i**3))
    g_wall[i] = 2*csum_wall[i]/(nobs*N*N)*L**3/(4*np.pi*delta**3/3*((i + 1)**3-i**3))
    dens[i] = csum_wall[i]/(N*nobs*delta*L*L) *1000
    results.write(str(i*delta)+' '+str(g[i])+' ' +str(g_wall[i])+' ' + str(csum_wall[i]/(N*nobs))+' '+str(dens[i])+'\n')
    i = i + 1
    
    
results.close()

#results = file('y_pos_wall.txt',"w")
#results.write('r y_posl\n')   

#results2 = file('y_neg_wall.txt',"w")
#results2.write('r y_neg\n')   
#i = 0

#while i < numbins:
#    results2.write(str(-i*deltay)+' '+str(sum_negy[i]/nobs)+'\n')
#    results.write(str(i*deltay)+' '+str(sum_y[i]/nobs)+'\n')
#    i = i + 1
    
    
#results.close()


	